from .base.increasing_progress_schedule import IncreasingProgressSchedule

from .functional import linear


class LinearIncreasing(IncreasingProgressSchedule):
    def _get_progress(self, step, total_steps):
        return linear(step, total_steps)



from .base.decreasing_progress_schedule import DecreasingProgressSchedule

from .functional import linear


class LinearDecreasing(DecreasingProgressSchedule):
    def _get_progress(self, step, total_steps):
        return linear(step, total_steps)



from .base.schedule_base import ScheduleBase


class Constant(ScheduleBase):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return f"{type(self).__name__}(value={self.value})"

    def get_value(self, step, total_steps):
        self._check_steps(step, total_steps)
        return self.value



__version__ = "0.0.1"

# base
from .base.decreasing_progress_schedule import DecreasingProgressSchedule
from .base.increasing_progress_schedule import IncreasingProgressSchedule
from .base.schedule_base import ScheduleBase
#
from .constant import Constant
from .cosine_decreasing import CosineDecreasing
from .cosine_increasing import CosineIncreasing
from .linear_decreasing import LinearDecreasing
from .linear_increasing import LinearIncreasing



from .base.increasing_progress_schedule import IncreasingProgressSchedule
from .functional import cosine


class CosineIncreasing(IncreasingProgressSchedule):
    def _get_progress(self, step, total_steps):
        return cosine(step, total_steps)



from .base.decreasing_progress_schedule import DecreasingProgressSchedule
from .functional import cosine


class CosineDecreasing(DecreasingProgressSchedule):
    def _get_progress(self, step, total_steps):
        return cosine(step, total_steps)



from .schedule_base import ScheduleBase


class ProgressSchedule(ScheduleBase):
    def __init__(self, abs_start_value, abs_delta, exclude_first=False, exclude_last=False):
        self.abs_start_value = abs_start_value
        self.abs_delta = abs_delta
        self.exclude_first = exclude_first
        self.exclude_last = exclude_last

    def __str__(self):
        return (
            f"{type(self).__name__}("
            f"start={self.abs_start_value}, "
            f"end={self.abs_start_value + self.abs_delta}, "
            f"excl_first={self.exclude_first}, "
            f"excl_last={self.exclude_last}"
            f")"
        )

    def get_value(self, step, total_steps):
        self._check_steps(step, total_steps)
        if self.exclude_last:
            total_steps += 1
        if self.exclude_first:
            step += 1
            total_steps += 1
        # get progress of schedule (going from 0 to 1)
        progress = self._get_progress(step, total_steps)
        # adjust to "absolute value" (i.e. real learning rate)
        return self.abs_start_value + progress * self.abs_delta

    def _get_progress(self, step, total_steps):
        raise NotImplementedError



import math


def linear(step, total_steps):
    """ linearly increasing from [0 to 1] """
    return step / max(1, total_steps - 1)


def cosine(step, total_steps):
    """ cosine schedule from [0 to 1] """
    progress = step / max(1, total_steps - 1)
    return 1 - (1 + math.cos(math.pi * progress)) / 2



from .progress_schedule import ProgressSchedule


class DecreasingProgressSchedule(ProgressSchedule):
    def __init__(self, abs_max_value=1., abs_end_value=0., **kwargs):
        delta = abs_end_value - abs_max_value
        assert delta <= 0.
        super().__init__(abs_start_value=abs_max_value, abs_delta=delta, **kwargs)

    def _get_progress(self, step, total_steps):
        raise NotImplementedError



class ScheduleBase:
    @staticmethod
    def _check_steps(step, total_steps):
        assert 0 <= step < total_steps, f"0 <= step < total_steps (step={step} total_steps={total_steps})"

    def get_value(self, step, total_steps):
        raise NotImplementedError

    def __repr__(self):
        return str(self)

    def __str__(self):
        raise NotImplementedError



from .progress_schedule import ProgressSchedule


class IncreasingProgressSchedule(ProgressSchedule):
    def __init__(self, abs_start_value=0., abs_max_value=1., **kwargs):
        delta = abs_max_value - abs_start_value
        assert delta >= 0.
        super().__init__(abs_start_value=abs_start_value, abs_delta=delta, **kwargs)

    def _get_progress(self, step, total_steps):
        raise NotImplementedError
